////////////////////////////////////////////////////////////////////////////////
#include "neighbourhood.h"
#include <stdio.h>
////////////////////////////////////////////////////////////////////////////////
void Neighbourhood::clearMasks()
{
    int i;

    for(i=0;i<cellNbMask.size();++i)
    {
        if(cellNbMask[i]) { delete[] cellNbMask[i]; cellNbMask[i] = 0; }
    }
    cellNbMask.resize(0);
}

void Neighbourhood::newMasks(int n)
{
    int i;

    clearMasks();
    cellNbMask.resize(CELLS);

    for(i=0;i<CELLS;++i)
    {
        cellNbMask[i] = new uint[n]; memset(cellNbMask[i],0,sizeof(uint)*n);
    }
}
////////////////////////////////////////////////////////////////////////////////
void Neighbourhood::init(vector<BrickRot>& br,int NB0)
{
    // define upto CNSBITSL neighbours for each cell
    allCellNeighbours();
    // resize (and clear) cell-neighbour-masks for large neighbourhood
    newMasks(NSUINTL);
    // setup cell-neighbour-masks for large neighbourhood
    initMasksLarge();
    // get a meassure for the usefullness of the neighbours ...
    getNbUsage(br,NB0);
    // ... and use it to reduce the number of neighbours to CNSBITS
    reduce(NB0);
    // resize (and clear) cell-neighbour-masks for final neighbourhood
    newMasks(NSUINT);
   // setup cell-neighbour-masks for final neighbourhood
    initMasks();
}
////////////////////////////////////////////////////////////////////////////////
void Neighbourhood::allCellNeighbours()
{
    int i, j, k, ic;

    cellNb.resize(0);
    cellNb.resize(CELLS);
    cellNbCoord.resize(0);
    cellNbCoord.resize(CELLS);

    for(i=0;i<NI;++i) for(j=0;j<NJ;++j) for(k=0;k<NK;++k)
    {
        ic = cellIndex(i,j,k);
        // face neighbours
        addCheckedNeighbour(ic,i-1,j,k);
        addCheckedNeighbour(ic,i+1,j,k);
        addCheckedNeighbour(ic,i,j-1,k);
        addCheckedNeighbour(ic,i,j+1,k);
        addCheckedNeighbour(ic,i,j,k-1);
        addCheckedNeighbour(ic,i,j,k+1);
        // edge neighbours
        addCheckedNeighbour(ic,i-1,j-1,k);
        addCheckedNeighbour(ic,i+1,j-1,k);
        addCheckedNeighbour(ic,i-1,j+1,k);
        addCheckedNeighbour(ic,i+1,j+1,k);
        addCheckedNeighbour(ic,i,j-1,k-1);
        addCheckedNeighbour(ic,i,j+1,k-1);
        addCheckedNeighbour(ic,i,j-1,k+1);
        addCheckedNeighbour(ic,i,j+1,k+1);
        addCheckedNeighbour(ic,i-1,j,k-1);
        addCheckedNeighbour(ic,i-1,j,k+1);
        addCheckedNeighbour(ic,i+1,j,k-1);
        addCheckedNeighbour(ic,i+1,j,k+1);
        // corner neighbours
        addCheckedNeighbour(ic,i-1,j-1,k-1);
        addCheckedNeighbour(ic,i+1,j-1,k-1);
        addCheckedNeighbour(ic,i-1,j+1,k-1);
        addCheckedNeighbour(ic,i+1,j+1,k-1);
        addCheckedNeighbour(ic,i-1,j-1,k+1);
        addCheckedNeighbour(ic,i+1,j-1,k+1);
        addCheckedNeighbour(ic,i-1,j+1,k+1);
        addCheckedNeighbour(ic,i+1,j+1,k+1);
    }
}

void Neighbourhood::addCheckedNeighbour(int ic,int i,int j,int k)
{
    if(cellNb[ic].size()<CNSBITSL && i>=0 && i<NI && j>=0 && j<NJ && k>=0 && k<NK)
    {
        cellNb[ic].push_back(cellIndex(i,j,k));
        cellNbCoord[ic].push_back(Coord(i,j,k));
    }
}
////////////////////////////////////////////////////////////////////////////////
void Neighbourhood::initMasksLarge()
{
    // build the large neighbour-states of all cells

    int ic, ic1, in, i, j, k, ii, jj, kk;

    for(i=0;i<NI;++i) for(j=0;j<NJ;++j) for(k=0;k<NK;++k)
    {
        ic = cellIndex(i,j,k);
        for(ii=-1;ii<2;++ii) for(jj=-1;jj<2;++jj) for(kk=-1;kk<2;++kk)
        {
            ic1 = checkedIndex(ic,i+ii,j+jj,k+kk);
            if(ic1!=ic)
            {
                // check, if ic1 recognizes ic as neighbour
                for(in=0;in<cellNb[ic1].size();++in)
                {
                    // if so, setting ic will set this neighbour of cell ic
                    if(cellNb[ic1][in]==ic) setLarge(cellNbMask[ic],ic1,in);
                }
            }
        }
    }
}

int Neighbourhood::checkedIndex(int ifalse,int i,int j,int k) const
{
    if(i<0 || i>=NI || j<0 || j>=NJ || k<0 || k>=NK) return ifalse;

    return cellIndex(i,j,k);
}

void Neighbourhood::setLarge(uint *p,int ic,int in)
{
    // set neighbour in of cell ic

    int layer(ic/UINTCNSL);
    int entry(ic-layer*UINTCNSL);

    p[layer] |= 1<<(in+entry*CNSBITSL);
}
////////////////////////////////////////////////////////////////////////////////
void Neighbourhood::getNbUsage(vector<BrickRot>& br,int NB0)
{
    int ib, ibr, ibc, nbr, nbc, i, j, k, ic;
    uint p[NSUINTL];
    Coord shift;
    Brick brick;

    value.resize(NSTATESL); value.assign(NSTATESL,0);

    for(ib=NB0;ib<br.size();++ib)   // only bricks used for cell-filling
    {
        nbc = br[ib].nCell;         // number of cells of ib
        nbr = br[ib].bricks.size(); // number of orientations of ib
        for(ibr=0;ibr<nbr;++ibr)
        {
            // perform every possible shift
            shift = br[ib].nijk[ibr];
            for(k=0;k<shift.k;++k) for(i=0;i<shift.i;++i) for(j=0;j<shift.j;++j)
            {
                brick = br[ib].bricks[ibr];
                brick.shift(Coord(i,j,k));

                // build the neighbour-mask of the resulting move
                memset(p,0,NSBYTEL);
                for(ibc=0;ibc<nbc;++ibc)
                {
                    bmSet(p,cellNbMask[cellIndex(brick.cells[ibc])],NSUINTL);
                }

                // nonzero bits indicate a recognized neighbouring brick-cell
                for(ibc=0;ibc<nbc;++ibc)
                {
                    ic = cellIndex(brick.cells[ibc]);
                    ++value[(ic<<CNSBITSL)|getLarge(p,ic)];
                }
            }
        }
    }
}

uint Neighbourhood::getLarge(uint *p,int ic) const
{
    // return cell-neighbour-state of cell ic

    int layer(ic/UINTCNSL);
    int entry(ic-layer*UINTCNSL);
    
    return ( p[layer] >> entry*CNSBITSL ) & CNSMASKL;
}
////////////////////////////////////////////////////////////////////////////////
void Neighbourhood::reduce(int NB0)
{
    int ic, iremove, nremove, in, in0, length, cpos, is, i, ib;
    vector<float> nUse(CNSBITSL);
    int n;
    float pfree, mx, mn, powp[CNSBITSL+1];
    vector<int> ncount(CNSTATESL);

    // average probability of finding i-1 cells free
    powp[0] = 0;
    for(i=1;i<=CNSBITSL;++i)
    {
        powp[i] = 0;
        for(ib=NB0;ib<NB;++ib)
        {
            pfree = float(NB-ib)/NB;
            powp[i] += (1-pfree)*pow(pfree,i-1);
        }
        powp[i] /= NB-NB0;
    }

    // number of non-zero bits of a cell-neighbour-state is
    for(in=0;in<CNSBITSL;++in)
    {
        length = 1<<in;
        is = 0;
        while(is<CNSTATESL) // increment ncount where in is set
        {
            is += length; for(i=0;i<length;++i) ++ncount[is++];
        }
    }

    for(ic=0;ic<CELLS;++ic)
    {
        cpos = ic<<CNSBITSL;
        nremove = cellNb[ic].size()-CNSBITS;
        mx = 0;
        for(in=0;in<cellNb[ic].size();++in)
        {
            // count, how often it will be usefull, to check this neighbour
            nUse[in] = 0;
            length = 1<<in;
            is = 0;
            while(is<CNSTATESL)
            {
                is += length;
                for(i=0;i<length;++i)
                {
                    nUse[in] += powp[ncount[is]]*value[cpos|is];
                    ++is;
                }
            }
            if(nUse[in]>mx) mx = nUse[in];
        }
        // forget the most useless neighbours
        for(iremove=0;iremove<nremove;++iremove)
        {
            mn = mx;
            in0 = 0;
            for(in=0;in<cellNb[ic].size();++in)
            {
                if(nUse[in]<=mn) { mn = nUse[in]; in0 = in; }
            }
            n = cellNb[ic].size()-1;
            nUse[in0] = nUse[n];
            cellNb[ic][in0] = cellNb[ic][n];
            cellNb[ic].pop_back();
            cellNbCoord[ic][in0] = cellNbCoord[ic][n];
            cellNbCoord[ic].pop_back();
        }
        // sorting seems to improve caching-performance ???
        n = cellNb[ic].size();
        for(in=0;in<n-1;++in) for(i=in+1;i<n;++i)
        {
            if(nUse[in]>nUse[i])
            {
                swap(cellNb[ic][i],cellNb[ic][in]);
                swap(cellNbCoord[ic][i],cellNbCoord[ic][in]);
                swap(nUse[i],nUse[in]);
            }
        }
    }
}
////////////////////////////////////////////////////////////////////////////////
void Neighbourhood::initMasks()
{
    // build the neighbour-states of all cells

    int ic, ic1, in, i, j, k, ii, jj, kk;

    for(i=0;i<NI;++i) for(j=0;j<NJ;++j) for(k=0;k<NK;++k)
    {
        ic = cellIndex(i,j,k);
        for(ii=-1;ii<2;++ii) for(jj=-1;jj<2;++jj) for(kk=-1;kk<2;++kk)
        {
            ic1 = checkedIndex(ic,i+ii,j+jj,k+kk);
            if(ic1!=ic)
                for(in=0;in<cellNb[ic1].size();++in)
                    if(cellNb[ic1][in]==ic) set(cellNbMask[ic],ic1,in);
        }
    }
}

void Neighbourhood::set(uint *p,int ic,int in)
{
    // set neighbour in of cell ic

    int layer(ic/UINTCNS);
    int entry(ic-layer*UINTCNS);

    p[layer] |= 1 << (in+entry*CNSBITS);
}

void Neighbourhood::brickNbMask(const Brick& b,uint *p,vector<uint>& v) const
{
    // build the neighbour-state of brick b

    int i, nbc(b.cells.size()), ic, nout, ii;
    vector<uint> v1;
    Block block;

    memset(p,0,NSBYTE);
    v.resize(0);

    for(i=0;i<nbc;++i) bmSet(p,cellNbMask[cellIndex(b.cells[i])],NSUINT);

    // additional information
    block = b.asBlock();
    for(i=0;i<nbc;++i)
    {
        ic = cellIndex(b.cells[i]);

        // count, how many brick-cells are outside the neighbourhood of ic
        nout = nbc-1;
        for(ii=0;ii<cellNb[ic].size();++ii)
        {
            if(block.hit(cellNbCoord[ic][ii])) --nout;
        }

        // get all cell-neighbour-states, where this move can be performed
        v1.resize(0);
        findIndicees(0,get(p,ic),1,v1,1<<cellNb[ic].size());

        // store state-indicees and nout
        ic <<= CNSBITS;
        for(ii=0;ii<v1.size();++ii) v.push_back((ic|v1[ii])<<BCBITS | nout);
    }
}

void Neighbourhood::brickNbMask(const Brick& b,uint *p) const
{
    // build the neighbour-state of brick b (ib<NB0)

    int i, nbc(b.cells.size());

    memset(p,0,NSBYTE);

    for(i=0;i<nbc;++i) bmSet(p,cellNbMask[cellIndex(b.cells[i])],NSUINT);
}

uint Neighbourhood::get(uint *p,int ic) const
{
    // return cell-neighbour-state of cell ic

    int layer(ic/UINTCNS);
    int entry(ic-layer*UINTCNS);
    
    return ( p[layer] >> entry*CNSBITS ) & CNSMASK;
}

void Neighbourhood::findIndicees(uint m,uint m0,int pos,vector<uint>& v,int end) const
{
    // push all cell-neighbour-states that do not hit state m0

    v.push_back(m);
    while(pos!=end)
    {
        if((m0&pos)==0) findIndicees(m|pos,m0,pos<<1,v,end);
        pos <<= 1;
    }
}
////////////////////////////////////////////////////////////////////////////////
