////////////////////////////////////////////////////////////////////////////////
#include "puzzle.h"
#include <limits>
#include <stdio.h>
////////////////////////////////////////////////////////////////////////////////
Puzzle::Puzzle(): smoves(0), bmoves(0), pmoves(0), moves(0)
{
}

void Puzzle::clear()
{
    if(smoves) delete[] smoves; smoves = 0;
    if(bmoves) delete[] bmoves; bmoves = 0;
    if(pmoves) delete[] pmoves; pmoves = 0;
    if(moves) delete[] moves; moves = 0;
}

void Puzzle::solve(vector<Brick>& b)
{
    if(b.size()!=NB) { sizeMessage(); return; }

    timer.reset();
    timer.start();

    clear();

    initMoves(b);

    printf("init time: %e\n",timer.total());
    timer.start();

    hitCount = 0;
    memset(state,0,SBYTE);
    setBrick0();

    timer.stop();
    printf("total hits: %d\n",hitCount);
    printf("total time: %f (%f)\n",timer.total(),hitCount/timer.total());
}
// for sorting only
struct AuxBrick
{
    AuxBrick() {}
    AuxBrick(const Brick& bb,float xx): b(bb), x(xx) {}
    bool operator<(const AuxBrick& rhs) const { return x<rhs.x; }
    Brick b;
    float x;
};
void Puzzle::initMoves(vector<Brick>& b)
{
    int ib, is, ibc, ibr, nbr, ii, iii, icurr, n, i, j, k;
    vector<BrickRot> br;
    vector<uint> v;
    Brick brick;
    Coord shift;
    uint *p;
    Neighbourhood neighbourhood;
    vector<vector<vector<uint*> > > pm(NB-NB0);
    VVVI out(NB-NB0);
    vector<AuxBrick> vx;
    
    // number of moves with ibc cells are outside the cell-neighbourhood
    for(ib=0;ib<NB-NB0;++ib)
    {
        out[ib].resize(MXBC-1);
        for(ibc=0;ibc<MXBC-1;++ibc) out[ib][ibc].resize(NSTATES);
    }

    n = initBrickRot(b,br);
    moves = new uint[SUINT*n];
    bmoves = new const uint*[NB0+1];

    neighbourhood.init(br,NB0);

    // for bricks 0..NB0-1 all possible combinations will be checked
    icurr = 0;
    for(ib=0;ib<NB0;++ib)
    {
        bmoves[ib] = moves+icurr; // first move for ib; end-pointer for ib-1

        nbr = br[ib].bricks.size();
        for(ibr=0;ibr<nbr;++ibr)
        {
            shift = br[ib].nijk[ibr];
            for(k=0;k<shift.k;++k) for(i=0;i<shift.i;++i) for(j=0;j<shift.j;++j)
            {
                brick = br[ib].bricks[ibr];
                brick.shift(Coord(i,j,k));
                // store cell- and neighbour-state starting at move[icurr]
                p = moves+icurr; icurr += SUINT;
                memcpy(p,brick.asBlock().m,CSBYTE);
                neighbourhood.brickNbMask(brick,p+CSUINT);
            }
        }
    }
    bmoves[ib] = moves+icurr; // end pointer for ib==NB0-1

    // bricks NB0..NB-1 are used to fill cells
    n = 0;
    for(ib=0;ib<NB-NB0;++ib)
    {
        nbr = br[ib+NB0].bricks.size();

        // push Bricks for all moves of ib into vx
        vx.resize(0);
        for(ibr=0;ibr<nbr;++ibr)
        {
            shift = br[ib+NB0].nijk[ibr];
            for(k=0;k<shift.k;++k) for(i=0;i<shift.i;++i) for(j=0;j<shift.j;++j)
            {
                brick = br[ib+NB0].bricks[ibr];
                brick.shift(Coord(i,j,k));
                vx.push_back(AuxBrick(brick,brick.minCellIndex()));
            }
        }
        // sorting tries to minimize distance between moves filling a given cell
        sort(vx.begin(),vx.end());

        // pm[ib][is] stores pointers to the moves of ib
        pm[ib].resize(NSTATES);
        for(ii=0;ii<vx.size();++ii)
        {
            // store cell- and neighbour-state starting at move[icurr]
            p = moves+icurr; icurr += SUINT;
            brick = vx[ii].b;
            memcpy(p,brick.asBlock().m,CSBYTE);
            neighbourhood.brickNbMask(brick,p+CSUINT,v);
            for(iii=0;iii<v.size();++iii)
            {
                is = v[iii]>>BCBITS; // move is applicable for state is
                ibc = v[iii]&BCMASK; // move has ibc cells outside the checked neighbourhood
                ++out[ib][ibc][is];
                pm[ib][is].push_back(p);
            }
        }
        for(is=0;is<NSTATES;++is) n += pm[ib][is].size();
    }
    printf("total number of neighbour-state entries %d\n",n);

    pmoves = new const uint*[n]; // pointers to moves
    smoves = new const uint**[(NB-NB0)*NSTATES+1]; // state-moves (ib=NB0..NB-1)

    // fill the pointers collected in pm into one long array
    icurr = 0;
    for(is=0;is<NSTATES;++is)
    {
        for(ib=0;ib<NB-NB0;++ib)
        {
            smoves[is*(NB-NB0)+ib] = pmoves+icurr; // start of moves for (ib,is)
            for(ii=0;ii<pm[ib][is].size();++ii) pmoves[icurr++] = pm[ib][is][ii];
        }
    }
    smoves[is*(NB-NB0)] = pmoves+icurr;

    // calculate the values used to choose the next cell
    initValue(out);
}

int Puzzle::initBrickRot(vector<Brick>& b,vector<BrickRot>& br)
{
    int i, j(-1), ntot(0);
    float v, v0(-numeric_limits<float>::max());

    br.resize(NB);

    // get all brick orientations and choose brick to remove symmetries
    for(i=0;i<NB;++i)
    {
        br[i].init(i,b[i]); if(br[i].qualify(v) && v>v0) { j = i; v0 = v; }
    }
    if(j<0) symmetryMessage(); else br[j].removeGameSym();

    // bricks with fewer moves first
    sort(br.begin(),br.end());

    // some information
    for(i=0;i<NB;++i)
    {
        ntot += br[i].nMove; //statusMessage(br[i]);
    }
    printf("total number of moves %d\n",ntot);

    return ntot;
}

void Puzzle::initValue(VVVI& out)
{
    int ibc, ib, ic, in, is;
    float pfree, mx, powp[MXBC-1];
    vector<vector<float> > ev(NB-NB0);
    vector<float> value0(NSTATES);

    for(ib=0;ib<NB-NB0;++ib) ev[ib].resize(NSTATES);

    // average probability of finding ibc cells all free
    for(ibc=0;ibc<MXBC-1;++ibc)
    {
        powp[ibc] = 0;
        for(ib=NB0;ib<NB;++ib)
        {
            pfree = float(NB-ib)/NB;
            powp[ibc] += pow(pfree,ibc);
        }
        powp[ibc] /= NB-NB0;
    }

    // expectation value for reaching the next level
    for(ib=0;ib<NB-NB0;++ib) for(ic=0;ic<CELLS;++ic) for(in=0;in<CNSTATES;++in)
    {
        is = ic<<CNSBITS | in;
        for(ibc=0;ibc<MXBC-1;++ibc) ev[ib][is] += powp[ibc]*out[ib][ibc][is];
    }

    // get maximum of value0; minimum is zero
    mx = -1;
    for(is=0;is<NSTATES;++is)
    {
        value0[is] = 0;
        for(ib=0;ib<NB-NB0;++ib) value0[is] += ev[ib][is];
        if(value0[is]>mx) mx = value0[is];
    }
    // store value (12 bits) and corresponding state in an uint
    for(is=0;is<NSTATES;++is)
    {
        value0[is] = 1+value0[is]/mx*4094;
        value[is] = ( int(value0[is]) << NSBITS ) | is;
    }
}
////////////////////////////////////////////////////////////////////////////////
// the following code needs to be adjusted for different #cells, #neighbours...
////////////////////////////////////////////////////////////////////////////////
void Puzzle::setBrick0()
{
    const uint *p=bmoves[0], *pe=bmoves[1];

    for(;p!=pe;p+=SUINT) if((state[0]&p[0]|state[1]&p[1])==0)
    {
        bmSetState(state,p); setBrick1(); bmUnSetState(state,p);
    }
}
void Puzzle::setBrick1()
{
    const uint *p=bmoves[1], *pe=bmoves[2];

    for(;p!=pe;p+=SUINT) if((state[0]&p[0]|state[1]&p[1])==0)
    {
        bmSetState(state,p); setBrick2(); bmUnSetState(state,p);
    }
}
void Puzzle::setBrick2()
{
    const uint *p=bmoves[2], *pe=bmoves[3];

    for(;p!=pe;p+=SUINT) if((state[0]&p[0]|state[1]&p[1])==0)
    {
        bmSetState(state,p); setBrick3(); bmUnSetState(state,p);
    }
}
void Puzzle::setBrick3()
{
    const uint *p=bmoves[3], *pe=bmoves[4];

    for(;p!=pe;p+=SUINT) if((state[0]&p[0]|state[1]&p[1])==0)
    {
        bmSetState(state,p); fillCell(BMASK1); bmUnSetState(state,p);
    }
}

void Puzzle::fillCell4()
{
    const uint **pp, ***ppp=smoves+(NB-NB0)*chooseCell(value,state);

    for(pp=*ppp++;pp!=*ppp;++pp) if((state[0]&(*pp)[0]|state[1]&(*pp)[1])==0)
    {
        bmSetState(state,*pp); fillCell(BMASK1&~(1<<4)); bmUnSetState(state,*pp);
    }
    for(pp=*ppp++;pp!=*ppp;++pp) if((state[0]&(*pp)[0]|state[1]&(*pp)[1])==0)
    {
        bmSetState(state,*pp); fillCell(BMASK1&~(1<<5)); bmUnSetState(state,*pp);
    }
    for(pp=*ppp++;pp!=*ppp;++pp) if((state[0]&(*pp)[0]|state[1]&(*pp)[1])==0)
    {
        bmSetState(state,*pp); fillCell(BMASK1&~(1<<6)); bmUnSetState(state,*pp);
    }
    for(pp=*ppp++;pp!=*ppp;++pp) if((state[0]&(*pp)[0]|state[1]&(*pp)[1])==0)
    {
        bmSetState(state,*pp); fillCell(BMASK1&~(1<<7)); bmUnSetState(state,*pp);
    }
    for(pp=*ppp++;pp!=*ppp;++pp) if((state[0]&(*pp)[0]|state[1]&(*pp)[1])==0)
    {
        bmSetState(state,*pp); fillCell(BMASK1&~(1<<8)); bmUnSetState(state,*pp);
    }
    for(pp=*ppp++;pp!=*ppp;++pp) if((state[0]&(*pp)[0]|state[1]&(*pp)[1])==0)
    {
        bmSetState(state,*pp); fillCell(BMASK1&~(1<<9)); bmUnSetState(state,*pp);
    }
    for(pp=*ppp++;pp!=*ppp;++pp) if((state[0]&(*pp)[0]|state[1]&(*pp)[1])==0)
    {
        bmSetState(state,*pp); fillCell(BMASK1&~(1<<10)); bmUnSetState(state,*pp);
    }
    for(pp=*ppp++;pp!=*ppp;++pp) if((state[0]&(*pp)[0]|state[1]&(*pp)[1])==0)
    {
        bmSetState(state,*pp); fillCell(BMASK1&~(1<<11)); bmUnSetState(state,*pp);
    }
}

void Puzzle::fillCell(uint bstate)
{
    if(bstate==0)
    {
        if(++hitCount%5000==0) timeMessage();
    }
    else
    {
        // check all possibilities to fill the cell choosen by chooseCell()
        // bstate-bits indicate unused bricks
    
        const uint **pp, ***ppp=smoves+(NB-NB0)*chooseCell(value,state);

        if(bstate&MASK_4<<4)
        {
            pp = *ppp++;
            if(bstate&1<<4)
            {
                for(;pp!=*ppp;++pp) if((state[0]&(*pp)[0]|state[1]&(*pp)[1])==0)
                {
                    bmSetState(state,*pp);
                    fillCell(bstate&~(1<<4));
                    bmUnSetState(state,*pp);
                }
            }
            pp = *ppp++;
            if(bstate&1<<5)
            {
                for(;pp!=*ppp;++pp) if((state[0]&(*pp)[0]|state[1]&(*pp)[1])==0)
                {
                    bmSetState(state,*pp);
                    fillCell(bstate&~(1<<5));
                    bmUnSetState(state,*pp);
                }
            }
            pp = *ppp++;
            if(bstate&1<<6)
            {
                for(;pp!=*ppp;++pp) if((state[0]&(*pp)[0]|state[1]&(*pp)[1])==0)
                {
                    bmSetState(state,*pp);
                    fillCell(bstate&~(1<<6));
                    bmUnSetState(state,*pp);
                }
            }
            pp = *ppp++;
            if(bstate&1<<7)
            {
                for(;pp!=*ppp;++pp) if((state[0]&(*pp)[0]|state[1]&(*pp)[1])==0)
                {
                    bmSetState(state,*pp);
                    fillCell(bstate&~(1<<7));
                    bmUnSetState(state,*pp);
                }
            }
        }
        else
        {
            ppp += 4;
        }
        if(bstate&MASK_4<<8)
        {
            pp = *ppp++;
            if(bstate&1<<8)
            {
                for(;pp!=*ppp;++pp) if((state[0]&(*pp)[0]|state[1]&(*pp)[1])==0)
                {
                    bmSetState(state,*pp);
                    fillCell(bstate&~(1<<8));
                    bmUnSetState(state,*pp);
                }
            }
            pp = *ppp++;
            if(bstate&1<<9)
            {
                for(;pp!=*ppp;++pp) if((state[0]&(*pp)[0]|state[1]&(*pp)[1])==0)
                {
                    bmSetState(state,*pp);
                    fillCell(bstate&~(1<<9));
                    bmUnSetState(state,*pp);
                }
            }
            pp = *ppp++;
            if(bstate&1<<10)
            {
                for(;pp!=*ppp;++pp) if((state[0]&(*pp)[0]|state[1]&(*pp)[1])==0)
                {
                    bmSetState(state,*pp);
                    fillCell(bstate&~(1<<10));
                    bmUnSetState(state,*pp);
                }
            }
            pp = *ppp++;
            if(bstate&1<<11)
            {
                for(;pp!=*ppp;++pp) if((state[0]&(*pp)[0]|state[1]&(*pp)[1])==0)
                {
                    bmSetState(state,*pp);
                    fillCell(bstate&~(1<<11));
                    bmUnSetState(state,*pp);
                }
            }
        }
    }
}
////////////////////////////////////////////////////////////////////////////////
// CNSBITS==8 -> cell-neighbour-state can be accessed as byte
////////////////////////////////////////////////////////////////////////////////
#define CTESTL(ic) if( (ms&1<< ic    )&&pv[(ic<<8)+cnb[ic]]<v ) v=pv[(ic<<8)+cnb[ic]]
#define CTESTH(ic) if( (ms&1<<(ic-30))&&pv[(ic<<8)+cnb[ic]]<v ) v=pv[(ic<<8)+cnb[ic]]
////////////////////////////////////////////////////////////////////////////////
uint Puzzle::chooseCell(const uint *pv,const uint *ps)
{
    uint  ms(~ps[0]), v(1<<30);
    const unsigned char *cnb=(const unsigned char*)(ps+2);

    if(ms&MASK_6)
    {
        CTESTL(0);
        CTESTL(1);
        CTESTL(2);
        CTESTL(3);
        CTESTL(4);
        CTESTL(5);
    }
    if(ms&MASK_6<<6)
    {
        CTESTL(6);
        CTESTL(7);
        CTESTL(8);
        CTESTL(9);
        CTESTL(10);
        CTESTL(11);
    }
    if(ms&MASK_6<<12)
    {
        CTESTL(12);
        CTESTL(13);
        CTESTL(14);
        CTESTL(15);
        CTESTL(16);
        CTESTL(17);
    }
    if(ms&MASK_6<<18)
    {
        CTESTL(18);
        CTESTL(19);
        CTESTL(20);
        CTESTL(21);
        CTESTL(22);
        CTESTL(23);
    }
    if(ms&MASK_6<<24)
    {
        CTESTL(24);
        CTESTL(25);
        CTESTL(26);
        CTESTL(27);
        CTESTL(28);
        CTESTL(29);
    }

    ms = ~ps[1];

    if(ms&MASK_6)
    {
        CTESTH(30);
        CTESTH(31);
        CTESTH(32);
        CTESTH(33);
        CTESTH(34);
        CTESTH(35);
    }
    if(ms&MASK_6<<6)
    {
        CTESTH(36);
        CTESTH(37);
        CTESTH(38);
        CTESTH(39);
        CTESTH(40);
        CTESTH(41);
    }
    if(ms&MASK_6<<12)
    {
        CTESTH(42);
        CTESTH(43);
        CTESTH(44);
        CTESTH(45);
        CTESTH(46);
        CTESTH(47);
    }
    if(ms&MASK_6<<18)
    {
        CTESTH(48);
        CTESTH(49);
        CTESTH(50);
        CTESTH(51);
        CTESTH(52);
        CTESTH(53);
    }
    if(ms&MASK_6<<24)
    {
        CTESTH(54);
        CTESTH(55);
        CTESTH(56);
        CTESTH(57);
        CTESTH(58);
        CTESTH(59);
    }

    return v&NSMASK;
}
////////////////////////////////////////////////////////////////////////////////
void Puzzle::timeMessage() const
{
    printf("%d %f %f\n",hitCount,timer.total(),hitCount/timer.partial());
}

void Puzzle::sizeMessage() const
{
    printf("expected %d bricks\n",NB);
}

void Puzzle::symmetryMessage() const
{
    printf("all bricks have symmetries; will find too much solutions\n");
}

void Puzzle::statusMessage(const BrickRot& br) const
{
    printf("Brick %2d: rotations (unique) %2d (%2d) cells %2d faces %2d moves %3d\n",
            br.brickId,
            br.nRot,
            br.bricks.size(),
            br.nCell,
            br.nSurf,
            br.nMove);
}
////////////////////////////////////////////////////////////////////////////////
